#!/usr/bin/python
# -*- coding: iso-8859-15 -*-

## ----------------------------------
## TP2 
## ----------------------------------

import os
import graphviz

## ----------------------------------
## global variables and data structure

setNodes = []

# a list of adjacencies : {node: [[neighbor1,weight], [neighbor2,weight], ...], node2:...}
listAdj = {}

## ----------------------------------
## --- GRAPH CONSTRUCTION ---

##
## vertex creation

def createVertex(id):
    """insert the vertex in your data structure"""
    
    setNodes.append(id)
    listAdj[id] = []

##
## edge creation

def createEdge(node1, node2, value):
    """insert the edge linking two vertices 'node1' and 'node2' in your data structure
    
    don't forget to add the 'friendship value' !!!
    Here we store the opposite in order to use Prim and Kruskal in their genuine form
    and to use the minimum spanning tree
    """

    # put the value on INTEGER
    if node1 in setNodes and node2 in setNodes:
        listAdj[node1].append([node2, int(value)*(-1)])
        listAdj[node2].append([node1, int(value)*(-1)])

## ----------------------------------
## --- PART 1 ---
##
## create the cheapest broadcast tree


def cheapestTree(node="Babet"):
    """return the 'overall friendship cost' and the set of edges
    
    result is (an integer, a list of pair of nodes)
    """    
    
    ## we consider Kruskal algorithm:
    ## sort the edges by increasing value, then pick iteratively the edges
    ## without introducing cycles... until all nodes are picked
    
     # construct the list of all possible edges
    listEdges = []
    for node in setNodes:
        for (neighbor,value) in listAdj[node]:
            if [neighbor,node,value] not in listEdges:
                listEdges.append([node, neighbor, value])

    ## sort it by increasing value
    listEdges.sort(lambda x,y:x[2]-y[2])
    
    ## NB we could have used external python modules
    ## queue (https://docs.python.org/3.5/library/queue.html#queue.PriorityQueue)
    ## or heapq (https://docs.python.org/3.5/library/heapq.html#module-heapq). 

    ## initialisation
    alreadyInTree = []
    tree = []
    cost = 0
    
    ## iteration
    for [node1, node2, value] in listEdges:
        if not cycleInTree(node1, node2, tree):
            # this edge does not produce any cycle
            tree.append([node1, node2])
            if node1 not in alreadyInTree:
                alreadyInTree.append(node1)
            if node2 not in alreadyInTree:
                alreadyInTree.append(node2)
            cost += value

    return (cost, tree)
        
def cycleInTree(id1, id2, graph):
    """return true if there is exists a path between id1 and id2 in graph
    use the DFS algorithm to construct the connected component of id1 """
    
    alreadyVisitedNodes = []
    soonVisitedNodes = []

    soonVisitedNodes.append(id1)

    while soonVisitedNodes <> []:

        # pick the last node that has been inserted
        node = soonVisitedNodes.pop()
        alreadyVisitedNodes.append(node)

        ## construct the list of its neighbors
        neighborsList = []
        for (node1, node2) in graph:
            if node1 == node:
                neighborsList.append(node2)
            if node2 == node:
                neighborsList.append(node1)
                        
        ## each neighbor not yet visited will be visited soon
        for neighbor in neighborsList:
            if neighbor not in alreadyVisitedNodes:
                soonVisitedNodes.append(neighbor)

    ## return true if id2 is within this connected component
    return (id2 in alreadyVisitedNodes)


## ----------------------------------
## --- PART 2 ---
##
## find the better root
## we may use a centrality measure here (the closeness centrality)
## We may also develop our own method to find the root node inducing the
## the tree with the minimal depth.

def betterRoot(tree):
    """return the node so that the tree depth is minimal
    
    'tree' is a list of pair of nodes
    """
    # starting from the tree, we start from the leaves and we discard iteratively
    # all fathers until we reach one unique root or two possible roots.

    # 'savedNodes' is a set of nodes that could be the best root
    # Initially, they all could be
    savedNodes = setNodes

    # we can determine the best root when there is only one or two nodes in
    # this list 'savedNodes'
    while len(savedNodes)>2:
        # compute the node degree for the tree involving only the nodes
        # in the 'savedNodes' set.
        dictNode = giveDegree(tree, savedNodes)
        
        # rebuild savedNodes by picking only the nodes that are not leaf.
        savedNodes = []
        for (node,degree) in dictNode.iteritems():
            if degree > 1:
                savedNodes.append(node)    	       
    # there is one or two nodes in the 'savedNodes' list. Return one of them
    return savedNodes[0]       
    

def giveDegree(tree, inNodes):
    """work in the tree 'tree' induced by inNodes

    return a hashtable {node: degree}
    """

    listNodes = {}

    for (node1,node2) in tree:
        if node1 in inNodes and node2 in inNodes:
            if listNodes.has_key(node1):
                listNodes[node1] += 1
            else:
                listNodes[node1] = 1

            if listNodes.has_key(node2):
                listNodes[node2] += 1
            else:
                listNodes[node2] = 1

    return listNodes
    
    
## ----------------------------------
## ----------------------------------
## ----------------------------------
## --- MAIN ---

    
fileGraph = open("promotion.graph", 'r')
for line in fileGraph:
    
    lineContent = line.split(" ")

    ## retrieve id from file and create vertex
    if "id" in lineContent:
        ident = lineContent[1][:-1]
        createVertex(ident)
        
    ## retrieve two ids and a value from file and create edge between them
    if "edge" in lineContent:
        id1 = lineContent[1]
        id2 = lineContent[2]
        weight = lineContent[3][:-1]
        createEdge(id1, id2, weight)

## test the PART 1
res = cheapestTree()
print
print "The 'overall cost' is", res[0]*(-1)
print "Here is the set of communication links to create: "
for edge in res[1]:
    print str(edge[0])+" --- "+str(edge[1])
graphviz.showcheapestTree(res) # show the tree: blue edges
    
## test the PART 2
print
print "The best root is", betterRoot(res[1])


